/* nppt.c - format input as paper tape in a variety of ways.  */
/* RCS: $Id: nppt.c,v 1.7 1998/12/19 23:30:38 jsm28 Exp $ */

/* Copyright (c) Joseph Samuel Myers 1996, 1997, 1998.  All rights reserved.
 *
 * This file is part of nppt, a program to format its input in a
 * representation of paper tape.
 *
 * Nppt is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Nppt is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with nppt; see the file COPYING.  If not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * As a special exception, if the PostScript code contained in this file
 * is included in an output file from this program, this does not by itself
 * cause the output file to be covered by the GNU General Public License.
 * This exception does not however invalidate any other reasons why the
 * output file might be covered by the GNU General Public License.
 *
 */

#include <ctype.h>
#include <errno.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "util.h"
#include "winsize.h"

#define VERSION	"nppt 0.1\n"

struct tape_operations {
  void (*start_tape)(void); /* Function to start the tape */
  void (*print_character)(unsigned int, int); /* Function to print
					       * one character */
  void (*end_tape)(void); /* Function to end the tape */
};

struct tape_type {
  const char *name; /* Name of the tape type */
  const char *desc; /* Description */
  unsigned int max; /* Max value punchable */
  struct tape_operations ops; /* Operations on the tape */
  const void *params; /* Parameters for these operations */
};

const struct tape_operations *tape;
const void *tape_params;
unsigned int tape_max;

/* Traditional vertical tape display support */

struct vert_tape_params {
  const char *top; /* Top bit */
  const char *noholes; /* Basic string for each row (no holes) */
  const char *holes; /* Basic string for each row (holes) */
  const char *bottom; /* Bottom bit */
  const size_t *hole_bits; /* Bit for each place */
  size_t text_place; /* Characters from right that the text can be displayed */
};

const size_t bsd_tape_bits[] = { 7, 6, 5, 4, 3, 2, 1, 0 };
const struct vert_tape_params bsd_tape_params = {
  "___________\n",
  "|     .   |\n",
  "|ooooo.ooo|\n",
  "___________\n",
  bsd_tape_bits,
  1
};

void vert_start_tape(void);
void vert_print_char(unsigned int, int);
void vert_end_tape(void);

/* Support for horizontal ASCII art tapes */

struct horiz_tape_params {
  size_t height; /* Height of tape */
  const char *const *left; /* Strings for each left column */
  const char *noholes; /* A column (no holes) */
  const char *holes; /* A column (holes) */
  const char *const *right; /* Right part */
  const size_t *hole_bits; /* Bit for each place */
  size_t text_place; /* Characters from bottom text can be displayed
		      * (maybe 0) */
};

const size_t horiz8_tape_bits[] = { 7, 6, 5, 4, 3, 2, 1, 0 };
const char *const horiz8_left[] = {
  "      <    ",
  "     /.\\   ",
  "    / . \\  ",
  "   /  .  \\ ",
  "  /   .   -",
  " /    .   -",
  "-     .   -",
  NULL
};
const char *const horiz8_right[] = {
  "-     .   -",
  "-     .   -",
  "-     <   -",
  "-    / \\  -",
  "-   /   \\ -",
  "-  /     \\-",
  "- /        ",
  "-/         ",
  NULL
};
const struct horiz_tape_params horiz8_tape_params = {
  11,
  horiz8_left,
  "-     .   -",
  "-OOOOO.OOO-",
  horiz8_right,
  horiz8_tape_bits,
  0
};

const size_t horiz7_tape_bits[] = { 6, 5, 4, 3, 2, 1, 0 };
const char *const horiz7_left[] = {
  "     <    ",
  "    /.\\   ",
  "   / . \\  ",
  "  /  .  \\ ",
  " /   .   -",
  "-    .   -",
  NULL
};
const char *const horiz7_right[] = {
  "-    .   -",
  "-    .   -",
  "-    <   -",
  "-   / \\  -",
  "-  /   \\ -",
  "- /     \\-",
  "-/        ",
  NULL
};
const struct horiz_tape_params horiz7_tape_params = {
  10,
  horiz7_left,
  "-    .   -",
  "-OOOO.OOO-",
  horiz7_right,
  horiz7_tape_bits,
  0
};

const size_t horiz6_tape_bits[] = { 5, 4, 3, 2, 1, 0 };
const char *const horiz6_left[] = {
  "    <    ",
  "   /.\\   ",
  "  / . \\  ",
  " /  .  \\ ",
  "-   .   -",
  NULL
};
const char *const horiz6_right[] = {
  "-   .   -",
  "-   .   -",
  "-   <   -",
  "-  / \\  -",
  "- /   \\ -",
  "-/     \\-",
  NULL
};
const struct horiz_tape_params horiz6_tape_params = {
  9,
  horiz6_left,
  "-   .   -",
  "-OOO.OOO-",
  horiz6_right,
  horiz6_tape_bits,
  0
};

const size_t horiz5_tape_bits[] = { 4, 3, 2, 1, 0 };
const char *const horiz5_left[] = {
  "   <    ",
  "  /.\\   ",
  " / . \\  ",
  "-  .  \\ ",
  "-  .   -",
  NULL
};
const char *const horiz5_right[] = {
  "-  .   -",
  "-  .   -",
  "-  <   -",
  "- / \\  -",
  "-/   \\ -",
  "      \\-",
  NULL
};
const struct horiz_tape_params horiz5_tape_params = {
  8,
  horiz5_left,
  "-  .   -",
  "-OO.OOO-",
  horiz5_right,
  horiz5_tape_bits,
  0
};

void horiz_start_tape(void);
void horiz_print_char(unsigned int, int);
void horiz_end_tape(void);
void horiz_add_str(const char *);

/* Support for PostScript output */

#define MM_TO_PT(len)	((len) * (72.0 / 25.4))
#define PT_TO_MM(len)	((len) * (25.4 / 72.0))
#define DEG_TO_RAD(angle)	((angle) * (3.14159265358979323846 / 180.0))

struct ps_tape_params {
  double tape_width; /* Tape width, in mm */
  double sprocket_height; /* Sprocket holes (centre), offset from
			   * reference (bottom) edge, in mm */
  int holes_below_sprocket; /* Number of holes below the sprocket hole */
  int character_bits; /* Total number of bits in a character */
  double track_spacing; /* Spacing of tracks (in mm) */
  double sprocket_diameter; /* Diameter of sprocket hole (in mm) */
  double data_diameter; /* Diameter of data hole (in mm) */
  double data_hole_offset; /* Offset of centre of data holes to the right of
			    * centre of sprocket holes (used on typesetter
			    * (TS) tape) (in mm) */
  double inter_character_space; /* Inter-character spacing, in mm */
  double upper_tearoff_angle; /* Angle of upper tearoff in radians
			       * (counterclockwise from the rightward
			       * horizontal) */
  double lower_tearoff_angle; /* Angle of lower tearoff in radians
			       * (clockwise from the rightward horizontal) */
  double line_thickness; /* Thickness of the lines used to draw the tape,
			  * in mm */
};

/* Parameters from Facit 4070 manual (full reference?????) */

const struct ps_tape_params ps8_tape_params = {
  25.40 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  8,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

const struct ps_tape_params ps7_tape_params = {
  22.23 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  7,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

const struct ps_tape_params ps6_tape_params = {
  22.23 /* +/- 0.08; this _is_ the same as for 7-track */,
  9.96 /* +/- 0.1 */,
  3,
  6,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

const struct ps_tape_params ps5_tape_params = {
  17.46 /* +/- 0.08 */,
  9.96 /* +/- 0.1 */,
  3,
  5,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.0,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

const struct ps_tape_params psTS_tape_params = {
  22.23 /* +/- 0.08 */,
  11.02 /* +/- 0.1 */,
  3,
  6,
  2.54 /* +/- 0.05 */,
  1.17 /* +0.05, -0.025 */,
  1.83 /* +/- 0.05 */,
  0.33 /* +/- 0.02; the effect of this offset is to align the _leading edge_
	* of the sprocket and data holes, not the centres */,
  2.54,
  DEG_TO_RAD(40.0),
  DEG_TO_RAD(40.0),
  0.35
};

void ps_start_tape(void);
void ps_print_char(unsigned int, int);
void ps_end_tape(void);

const struct tape_type tape_types[] = {
  { "BSD", "Traditional BSD/V7 ppt(6) format 8-track tape", 255U,
    { vert_start_tape, vert_print_char, vert_end_tape },
    &bsd_tape_params
  },
  { "horiz8", "Horizontal ASCII art 8-track tape", 255U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz8_tape_params
  },
  { "horiz7", "Horizontal ASCII art 7-track tape", 127U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz7_tape_params
  },
  { "horiz6", "Horizontal ASCII art 6-track tape", 63U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz6_tape_params
  },
  { "horiz5", "Horizontal ASCII art 5-track tape", 31U,
    { horiz_start_tape, horiz_print_char, horiz_end_tape },
    &horiz5_tape_params
  },
  { "PS8", "Standard 8-track tape (Encapsulated PostScript output)", 255U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps8_tape_params
  },
  { "PS7", "Standard 7-track tape (Encapsulated PostScript output)", 127U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps7_tape_params
  },
  { "PS6", "Standard 6-track tape (Encapsulated PostScript output)", 63U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps6_tape_params
  },
  { "PS5", "Standard 5-track tape (Encapsulated PostScript output)", 31U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &ps5_tape_params
  },
  { "PSTS", "6-track typesetter (TS) tape (Encapsulated PostScript output)",
    63U,
    { ps_start_tape, ps_print_char, ps_end_tape },
    &psTS_tape_params
  },
  { NULL, NULL, 0U,
    { NULL, NULL, NULL },
    NULL
  }
};

int main(int argc, char **argv); /* Main program */

void start_tape(void);
void print_cmd_arg(const char *arg);
void print_cmd_arg_space(void);
void print_stdin_char(unsigned int c);
void end_tape(void);

void print_tape_char(unsigned int c, int character);



int show_text_flag = 0; /* Show text alongside the tape?  */
unsigned long screen_columns; /* Number of screen columns (if needed) */
int override_screen_columns_flag = 0; /* Override the number of screen
					 columns from the command line? */

/* main - main program */
int
main(int argc, char **argv)
{
  const char *tape_type = NULL;
  /* Get the program name for error and usage messages */
  get_program_name(argv[0]);
  /* Parse options */
  while (1) {
#ifdef HAVE_GETOPT_LONG
    const struct option long_options[] = {
      { "help", no_argument, (int *)NULL, 'h' },
      { "version", no_argument, (int *)NULL, 'v' },
      { "tape-type", required_argument, (int *)NULL, 't' },
      { "show-text", no_argument, (int *)NULL, 'T' },
      { "columns", required_argument, (int *)NULL, 'c' },
      { NULL, 0, NULL, 0 }
    };
    int option_index;
#endif
    const char short_options[] = "hvt:Tc:";
    int ret;
#ifdef HAVE_GETOPT_LONG
    ret = getopt_long(argc, argv, short_options, long_options, &option_index);
#else
    ret = getopt(argc, argv, short_options);
#endif
    if (ret == -1)
      break; /* End of options */
    switch (ret) {
    case 'h':
      /* Help */
      xprintf("Usage: %s [ --help | --version ] [ --tape-type=TYPE]\n"
	      "\t[ --show-text ] [ --columns=NUMBER ] [ WORDS ]\n",
	      program_name);
      xprintf("\nRecognised tape types:\n\n");
      {
	const struct tape_type *i;
	for (i = tape_types; i->name != NULL; i++)
	  xprintf("%s\t%s\n", i->name, i->desc);
      }
      xflush(stdout, "standard output");
      exit(EXIT_SUCCESS);
    case 'v':
      /* Version */
      xprintf(VERSION);
      xflush(stdout, "standard output");
      exit(EXIT_SUCCESS);
    case 't':
      if (tape_type == NULL)
	tape_type = optarg;
      else
	die("multiple --tape-type or -t options");
      break;
    case 'T':
      if (show_text_flag)
	die("multiple --show-text or -T options");
      else
	show_text_flag = 1;
      break;
    case 'c':
      if (override_screen_columns_flag) {
	die("multiple --columns or -c options");
      } else {
	char *endptr;
	override_screen_columns_flag = 1;
	if (strtol(optarg, NULL, 0) < 0)
	  die("a negative number of columns does not make sense");
	errno = 0;
	screen_columns = strtoul(optarg, &endptr, 0);
	if (errno != 0)
	  die("--columns or -c value `%s': %s", optarg, strerror(errno));
	if ((endptr == optarg) || (*endptr != '\0'))
	  die("--columns or -c value `%s' is not numeric or has trailing junk",
	      optarg);
      }
      break;
    case '?':
      /* Bad option - error message printed */
      exit(EXIT_FAILURE);
    default:
      die("panic: bad return %d from getopt_long", ret);
    }
  }
  /* Now determine the type of tape display wanted.  */
  if (tape_type == NULL)
    tape_type = "BSD";
  {
    const struct tape_type *i;
    for (i = tape_types; i->name != NULL; i++) {
      if (!strcmp(i->name, tape_type)) {
	tape = &i->ops;
	tape_params = i->params;
	tape_max = i->max;
	break;
      }
    }
    if (i->name == NULL)
      die("tape type `%s' not known", tape_type);
  }
  /* If we have any non-option arguments then we print those as paper tape;
   * otherwise we read standard input.  */
  if (optind < argc) {
    int i;
    start_tape();
    for (i = optind; i < argc; i++) {
      print_cmd_arg(argv[i]);
      if (i < argc - 1)
	print_cmd_arg_space();
    }
    end_tape();
  } else {
    start_tape();
    while (1) {
      int c;
#ifdef HAVE_GETC_UNLOCKED
      c = getc_unlocked(stdin);
#else
      c = getc(stdin);
#endif
      if (c == EOF) {
	if (ferror(stdin))
	  die("error reading stdin: %s", strerror(errno));
	else {
	  /* EOF */
	  end_tape();
	  break;
	}
      }
      print_stdin_char(c);
    }
  }
  exit(EXIT_SUCCESS);
}

/* start_tape - start printing the tape */
void
start_tape(void)
{
  tape->start_tape();
}

/* print_cmd_arg - print something on the tape from a command line argument.
 * Maybe do character set translation.  */
void
print_cmd_arg(const char *arg)
{
  const char *c;
  for (c = arg; *c; c++)
    print_tape_char((unsigned char)*c, (unsigned char)*c);
}

/* print_cmd_arg_space - print a space corresponding to the gap between
 * command line arguments.  Maybe do character set translation.  */
void
print_cmd_arg_space(void)
{
  print_tape_char((unsigned char)' ', (unsigned char)' ');
}

/* print_stdin_char - print a character from stdin.  Maybe do character set
 * translation.  */
void
print_stdin_char(unsigned int c)
{
  print_tape_char(c, c);
}

/* end_tape - end printing the tape */
void
end_tape(void)
{
  tape->end_tape();
}

/* print_tape_char - print a character on the tape.  Here c is the raw
 * character for the tape; translations may have been done previously.
 * `character' is the character c represents.  */
void
print_tape_char(unsigned int c, int character)
{
  if (c > tape_max)
    die("can't print character %u, too wide for tape", c);
  tape->print_character(c, character);
}



/* Vertical tape display functions */

const struct vert_tape_params *vtp;
size_t vtlen;

/* vert_start_tape - start printing the tape, vertical ASCII art */
void
vert_start_tape(void)
{
  vtp = (const struct vert_tape_params *)tape_params;
  vtlen = strlen(vtp->noholes);
  xprintf("%s", vtp->top);
}

/* vert_print_char - print a character, vertical ASCII art */
void
vert_print_char(unsigned int c, int character)
{
  size_t p, q;
  q = 0;
  for (p = 0; p < vtlen; p++) {
    int c1, c2, pc;
    if (show_text_flag && p == (vtlen - vtp->text_place)) {
      int pc2;
      pc2 = (isprint(character) ? character : (unsigned char)' ');
      xputc(pc2, stdout, "standard output");
    }
    c1 = (unsigned char)vtp->noholes[p];
    c2 = (unsigned char)vtp->holes[p];
    if (c1 == c2) {
      pc = c1;
    } else {
      pc = ((c & (1 << vtp->hole_bits[q++])) ? c2 : c1);
    }
    xputc(pc, stdout, "standard output");
  }
}

/* vert_end_tape - end printing the tape, vertical ASCII art */
void
vert_end_tape(void)
{
  xprintf("%s", vtp->bottom);
  xflush(stdout, "standard output");
}



/* Horizontal tape display functions */

const struct horiz_tape_params *htp;
char *horiz_current;
char **horiz_buffers;
size_t horiz_buffers_allocated;
size_t horiz_buffers_pos;

/* horiz_start_tape - start printing the tape, horizontal ASCII art */
void
horiz_start_tape(void)
{
  const char *const *i;
  htp = (const struct horiz_tape_params *)tape_params;
  if (!override_screen_columns_flag) {
    struct window_size win;
    get_window_size(&win);
    screen_columns = win.cols;
  }
  /* Allocate buffers */
  {
    size_t amt;
    size_t j;
    amt = (screen_columns ? (screen_columns + 2) : 1024);
    horiz_buffers = xmalloc((htp->height + 1) * sizeof(char *));
    horiz_current = xmalloc(htp->height + 2);
    for (j = 0; j < htp->height + 1; j++)
      horiz_buffers[j] = xmalloc(amt);
    horiz_buffers_allocated = amt;
    horiz_buffers_pos = 0;
  }
  for (i = htp->left; *i != NULL; i++)
    horiz_add_str(*i);
}

/* horiz_print_char - print a character, horizontal ASCII art */
void
horiz_print_char(unsigned int c, int character)
{
  size_t p, q;
  char *hcp;
  q = 0;
  hcp = horiz_current;
  for (p = 0; p <= htp->height; p++) {
    char c1, c2;
    if (show_text_flag && p == (htp->height - htp->text_place)) {
      int pc2;
      pc2 = (isprint(character) ? character : (unsigned char)' ');
      *(hcp++) = pc2;
    }
    if (p == htp->height) {
      *(hcp++) = '\0';
      break;
    }
    c1 = htp->noholes[p];
    c2 = htp->holes[p];
    if (c1 == c2) {
      *(hcp++) = c1;
    } else {
      *(hcp++) = ((c & (1 << htp->hole_bits[q++])) ? c2 : c1);
    }
  }
  horiz_add_str(horiz_current);
}

/* horiz_end_tape - end printing the tape, horizontal ASCII art */
void
horiz_end_tape(void)
{
  const char *const *i;
  for (i = htp->right; *i != NULL; i++)
    horiz_add_str(*i);
  if (horiz_buffers_pos > 0) {
    size_t j;
    for (j = 0; j < htp->height + 1; j++) {
      horiz_buffers[j][horiz_buffers_pos] = '\n';
      horiz_buffers[j][horiz_buffers_pos + 1] = '\0';
      xprintf("%s", horiz_buffers[j]);
    }
  }
  xflush(stdout, "standard output");
  /* Things will get freed when the program exits */
}

/* horiz_add_str - add a string to a horizontal tape */
void
horiz_add_str(const char *str)
{
  size_t i, j;
  for (i = 0; str[i] != '\0'; i++) {
    horiz_buffers[i][horiz_buffers_pos] = str[i];
  }
  for (j = i; j < htp->height + 1; j++)
    horiz_buffers[j][horiz_buffers_pos] = ' ';
  horiz_buffers_pos++;
  if (horiz_buffers_pos == horiz_buffers_allocated - 2) {
    if (screen_columns) {
      /* Print this part of the tape */
      size_t k;
      for (k = 0; k < htp->height + 1; k++) {
	horiz_buffers[k][horiz_buffers_pos] = '\n';
	horiz_buffers[k][horiz_buffers_pos + 1] = '\0';
	xprintf("%s", horiz_buffers[k]);
      }
      if (show_text_flag)
	xprintf("\n");
      horiz_buffers_pos = 0;
    } else {
      /* Reallocate */
      size_t k;
      horiz_buffers_allocated *= 2;
      for (k = 0; k < htp->height + 1; k++)
	horiz_buffers[k] = xrealloc(horiz_buffers[k], horiz_buffers_allocated);
    }
  }
}



/* PostScript tape display functions */

const struct ps_tape_params *pstp;
size_t ps_number_of_chars;
size_t ps_chars_allocated;
unsigned int *ps_bits;
unsigned char *ps_chars;

/* ps_start_tape - start printing the tape, PostScript */
void
ps_start_tape(void)
{
  /* In order to know the bounding box we need to have all the tape data;
   * and it isn't clear whether EPSF files can use (atend) for the bounding
   * box; so we defer all the work to the end.  */
  pstp = (const struct ps_tape_params *)tape_params;
  ps_chars_allocated = 1024;
  ps_bits = xmalloc(ps_chars_allocated * sizeof(unsigned int));
  if (show_text_flag)
    ps_chars = xmalloc(ps_chars_allocated);
  ps_number_of_chars = 0;
}

/* ps_print_char - print a character, PostScript */
void
ps_print_char(unsigned int c, int character)
{
  /* Actually with PostScript output we just store the character (both
   * the bit pattern and the original character, if appropriate), to output
   * at the end.  */
  if (ps_number_of_chars == ps_chars_allocated) {
    ps_chars_allocated *= 2;
    ps_bits = xrealloc(ps_bits, ps_chars_allocated * sizeof(unsigned int));
    if (show_text_flag)
      ps_chars = xrealloc(ps_chars, ps_chars_allocated);
  }
  if (show_text_flag)
    ps_chars[ps_number_of_chars] = character;
  ps_bits[ps_number_of_chars++] = c;
}

/* ps_end_tape - end printing the tape, PostScript */
void ps_end_tape(void)
{
  /* This is where we do all the work */
#define MM_TO_UNITS(len)	((len) * (1000.0 / pstp->tape_width))
#define MIN(x, y)	(((x) < (y)) ? (x) : (y))
#define MAX(x, y)	(((x) > (y)) ? (x) : (y))
  double line_thickness; /* Line thickness, in the 1000 unit scale */
  double font_base; /* Base of the font, in the 1000 unit scale */
  double font_top; /* Top of the font, in the 1000 unit scale */
  double norm_left, norm_right, norm_width; /* dimensions of normal
					     * character */
  double lend_left, lend_right, lend_width; /* left end */
  int lend_char_positions, lend_left_pos;
  double rend_left, rend_right, rend_width; /* right end */
  int rend_char_positions, rend_left_pos;
  double sprocket_radius, data_radius; /* radii of holes */
  double data_hole_offset;
  double sprocket_height;
  double track_spacing;
  double utheta, dtheta;
  line_thickness = MM_TO_UNITS(pstp->line_thickness);
  font_base = 0.0 - (line_thickness / 2.0);
  font_top = 1000.0 + (line_thickness / 2.0);
  norm_width = MM_TO_UNITS(pstp->inter_character_space);
  sprocket_radius = MM_TO_UNITS(pstp->sprocket_diameter) / 2.0;
  data_radius = MM_TO_UNITS(pstp->data_diameter) / 2.0;
  data_hole_offset = MM_TO_UNITS(pstp->data_hole_offset);
  sprocket_height = MM_TO_UNITS(pstp->sprocket_height);
  track_spacing = MM_TO_UNITS(pstp->track_spacing);
  norm_left = MIN(0, MIN(norm_width / 2.0 - sprocket_radius,
			 norm_width / 2.0 + data_hole_offset - data_radius));
  norm_right = MAX(norm_width,
		   MAX(norm_width / 2.0 + sprocket_radius,
		       norm_width / 2.0 + data_hole_offset + data_radius));
  utheta = pstp->upper_tearoff_angle;
  dtheta = pstp->lower_tearoff_angle;
  /* The left end.  We work out the space needed from hole centre at the
   * left to the inner intersection of the lines at top and bottom.  If
   * holes from a character extend to the left of its strip, we adjust the
   * right boundary accordingly.  If they extend to the right, we adjust the
   * left boundary accordingly.  */
  {
    double tmpl, tmpr;
    double tmpu, tmpd;
    tmpu = (((1000.0 - sprocket_height) / tan(utheta))
	    + ((line_thickness / 2.0) * (1 / sin(utheta) - 1 / tan(utheta))));
    tmpd = ((sprocket_height / tan(dtheta))
	    + ((line_thickness / 2.0) * (1 / sin(dtheta) - 1 / tan(dtheta))));
    tmpr = MAX(tmpu, tmpd) - norm_left;
    tmpl = -(line_thickness / 2.0) * ((cos(utheta) + cos(dtheta))
				      / sin(utheta + dtheta));
    tmpl -= (norm_right - norm_width);
    lend_left_pos = -floor(tmpl / norm_width + 0.5);
    lend_char_positions = ceil(tmpr / norm_width + 0.5) + lend_left_pos;
    lend_left = 0;
    lend_right = lend_width = norm_width * lend_char_positions;
  }
  /* The right end - adjustments as commented above */
  {
    double tmpl, tmpr;
    double tmpu, tmpd;
    tmpu = (((1000.0 - sprocket_height) / tan(utheta))
	    + ((line_thickness / 2.0) * (1 / tan(utheta) + 1 / sin(utheta))));
    tmpd = ((sprocket_height / tan(dtheta))
	    + ((line_thickness / 2.0) * (1 / tan(dtheta) + 1 / sin(dtheta))));
    tmpr = MAX(tmpu, tmpd) - norm_left;
    tmpl = -(line_thickness / 2.0) * ((cos(utheta) + cos(dtheta))
				      / sin(utheta + dtheta));
    tmpl -= (norm_right - norm_width);
    rend_left_pos = -floor(tmpl/norm_width + 0.5);
    rend_char_positions = ceil(tmpr / norm_width + 0.5) + rend_left_pos;
    rend_left = 0;
    rend_right = rend_width = norm_width * rend_char_positions;
  }
  /* The EPSF header */
  {
    int llx, lly, urx, ury;
    time_t t;
    llx = 0;
    if (show_text_flag) {
      /* Courier characters can extend to 216/1000 below their box */
      lly = floor(-MM_TO_PT(pstp->inter_character_space) / 0.6 * 1.216);
    } else {
      lly = floor(-MM_TO_PT(pstp->line_thickness) / 2.0);
    }
    ury = ceil(72.0 + MM_TO_PT(pstp->line_thickness) / 2.0);
    urx = ceil((ps_number_of_chars + lend_char_positions + rend_char_positions)
	       * MM_TO_PT(pstp->inter_character_space));
    t = time(NULL);
    xprintf("%%!PS-Adobe-3.0 EPSF-3.0\n"
	    "%%%%BoundingBox: %d %d %d %d\n"
	    "%%%%LanguageLevel: 2\n"
	    "%%%%Creator: "
	    VERSION
	    "%%%%CreationDate: %s",
	    llx, lly, urx, ury, ctime(&t));
    if (show_text_flag)
      xprintf("%%%%DocumentNeededResources: font Courier\n");
    xprintf("\n\n");
  }
  /* The font */
  {
    int fontllx, fontlly, fonturx, fontury;
    int normllx, normurx;
    int r_urx, l_urx;
    /* This font was originally designed in 1996 for the Cambridge University
     * Computer Preservation Society
     * <URL:http://www.cam.ac.uk/CambUniv/Societies/cucps/>, whose logo
     * (see <URL: http://www.cam.ac.uk/CambUniv/Societies/cucps/logo.htm>)
     * consists of a representation of the Society's name on paper tape.
     * Some of the C comments below are derived from those in the original
     * PostScript version.
     *
     * The font defines the PaperTape8 (or PaperTape7, etc.) font, and the
     * procedures scaletape and (if appropriate) scaletext to scale the
     * fonts and set the current font.
     *
     * The method of using a font and printing strings is fundamentally
     * limited to 8-track tapes, but I don't know of wider versions.
     *
     * In 12pt Courier, characters are exactly(?) 0.1in across; this is
     * where the 0.6 scale factor used in various places comes from.  */
    xprintf("8 dict begin\n");
    /* The EPSF spec notes that we should initialise all variables before
     * binding a procedure that uses them, to avoid problems if the
     * including document uses the same names for operators.  Since this
     * font is included in EPSF files, which may in turn be included in
     * PostScript documents (originally CUCPS posters and flyers), we do
     * just this.  Note that this font definition as-is needs to go in the
     * script of an EPSF file, because of the extraneous things it defines.
     * Since we aren't using short names like x, problems aren't very likely,
     * but it is best to avoid them anyway.  */
    xprintf("/dotapestart 0 def /dotapeend 0 def /punchchar 0 def\n\n"
	    /* Used in BuildGlyph */);
    fontllx = floor(MIN(norm_left, MIN(lend_left, rend_left)));
    fontlly = floor(font_base);
    fonturx = ceil(MAX(norm_right, MAX(lend_right, rend_right)));
    fontury = ceil(font_top);
    normllx = floor(norm_left);
    normurx = ceil(norm_right);
    l_urx = ceil(lend_right);
    r_urx = ceil(rend_right);
    xprintf("8 dict begin\n"
	    "\n"
	    "/FontType 3 def\n"
	    "/FontMatrix [ .001 0 0 .001 0 0 ] def\n"
	    "/FontBBox [ %d %d %d %d ] def\n"
	    "\n",
	    fontlly, fontlly, fonturx, fontury);
    xprintf("/Encoding 256 array def\n"
	    "0 1 255 { Encoding exch dup 3 string cvs cvn put } for\n"
	    "\n");
    xprintf("/BuildChar {\n"
	    "  1 index /Encoding get exch get\n"
	    "  1 index /BuildGlyph get exec\n"
	    "} bind def\n"
	    "\n"
	    "/BuildGlyph {\n"
	    "  dup /.notdef eq\n"
	    "    { pop pop }\n"
	    "    { dup /tapestart eq\n"
	    "      { pop /CharProcs get begin dotapestart end }\n"
	    "      { dup /tapeend eq\n"
	    "        { pop /CharProcs get begin dotapeend end }\n"
	    "        %% Otherwise we assume it's a number\n"
	    "        { 3 string cvs cvi exch /CharProcs get begin\n"
	    "          punchchar end }\n"
	    "      ifelse }\n"
	    "    ifelse }\n"
	    "  ifelse\n"
	    "} bind def\n"
	    "\n");
    xprintf("/CharProcs 12 dict def\n"
	    "CharProcs begin\n"
	    "  /punchhole { newpath\n"
	    "               3 -1 roll %.6g add\n"
	    "               3 -1 roll %.6g mul %.6g add\n"
	    "               3 -1 roll %% x y r\n"
	    "               0 360 arc closepath fill\n"
	    "  } bind def\n",
	    norm_width / 2,
	    track_spacing,
	    sprocket_height - track_spacing * pstp->holes_below_sprocket);
    xprintf("  /sprocket { 0 %d %.6g punchhole } bind def\n"
	    "  /punchbit { dup %d ge { 1 add } if %.6g exch %.6g punchhole\n"
	    "  } bind def\n",
	    pstp->holes_below_sprocket,
	    sprocket_radius,
	    pstp->holes_below_sprocket,
	    data_hole_offset,
	    data_radius);
    xprintf("  /punchchar { %.6g 0 %d %d %d %d setcachedevice\n"
	    "               %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "               newpath\n"
	    "               0 0 moveto %.6g 0 lineto stroke\n"
	    "               0 1000 moveto %.6g 1000 lineto stroke\n"
	    "               sprocket\n"
	    "               1 0 1 %d { 3 copy pop and\n"
	    "                          0 gt { punchbit } { pop } ifelse\n"
	    "                          2 mul } for pop pop\n"
	    "  } bind def\n",
	    norm_width, normllx, fontlly, normurx, fontury,
	    line_thickness, norm_width, norm_width, pstp->character_bits - 1);
    xprintf("  /dotapestart { %.6g 0 0 %d %d %d setcachedevice\n"
	    "                 %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "                 newpath\n"
	    "                 %.6g 0 moveto %.6g 0 lineto %.6g %.6g lineto\n"
	    "                 %.6g 1000 lineto %.6g 1000 lineto\n"
	    "                 gsave stroke grestore %% draw edge\n"
	    "                 closepath clip\n"
	    "                 %d { sprocket %.6g 0 translate } repeat\n"
	    "  } bind def\n",
	    lend_width, fontlly, l_urx, fontury, line_thickness,
	    lend_right, ((norm_width / 2.0)
			 + (norm_width * lend_left_pos)
			 + sprocket_height / tan(dtheta)),
	    norm_width / 2.0 + norm_width * lend_left_pos,
	    sprocket_height,
	    ((norm_width / 2.0) + (norm_width * lend_left_pos)
	     + (1000.0 - sprocket_height) / tan(utheta)), lend_right,
	    lend_char_positions, norm_width);
    xprintf("  /dotapeend { %.6g 0 0 %d %d %d setcachedevice\n"
	    "               %.6g setlinewidth 0 setlinecap [ ] 0 setdash\n"
	    "               newpath\n"
	    "               0 0 moveto %.6g 0 lineto %.6g %.6g lineto\n"
	    "               %.6g 1000 lineto 0 1000 lineto\n"
	    "               gsave stroke grestore closepath clip\n"
	    "               %d { sprocket %.6g 0 translate } repeat\n"
	    "  } bind def\n",
	    rend_width, fontlly, r_urx, fontury, line_thickness,
	    ((norm_width / 2.0) + (norm_width * rend_left_pos)
	     + sprocket_height / tan(dtheta)),
	    norm_width / 2.0 + norm_width * rend_left_pos,
	    sprocket_height,
	    ((norm_width / 2.0) + (norm_width * rend_left_pos)
	     + (1000.0 - sprocket_height) / tan(utheta)),
	    rend_char_positions, norm_width);
    xprintf("end\n" /* CharProcs */
	    "\n"
	    "currentdict\n"
	    "end\n"
	    "/PaperTape%d exch definefont pop\n"
	    "\n",
	    pstp->character_bits);
    xprintf("/scaletape { /PaperTape%d findfont exch scalefont setfont\n"
	    "} bind def\n"
	    "\n"
	    "/scaletext { /Courier findfont exch scalefont /Courierfont\n"
	    "  exch def } bind def\n",
	    pstp->character_bits);
  }
  /* The actual message */
  {
    xprintf("%.6g scaletape\n", MM_TO_PT(pstp->tape_width));
    if (show_text_flag)
      xprintf("%.6g scaletext\n", MM_TO_PT(pstp->inter_character_space) / 0.6);
    xprintf("0 0 moveto /tapestart glyphshow\n");
    if (show_text_flag) {
      size_t i;
      xprintf("gsave Courierfont setfont\n"
	     "0 ( ) stringwidth pop 0.6 div neg rmoveto\n<");
      for (i = 0; i < ps_number_of_chars; i++) {
	xprintf("%02x", (unsigned int)ps_chars[i]);
	if (i % 32 == 31)
	  xprintf("\n");
      }
      xprintf("> show grestore\n");
    }
    {
      size_t i;
      xprintf("<");
      for (i = 0; i < ps_number_of_chars; i++) {
	xprintf("%02x", (unsigned int)ps_bits[i]);
	if (i % 32 == 31)
	  xprintf("\n");
      }
      xprintf("> show\n");
    }
    xprintf("/tapeend glyphshow\n"
	    "end\n"
	    "%%%%EOF");
  }
  xflush(stdout, "standard output");
}
